import QUB
import Commands
from Dialogs import *
from math import *
import os
import re
import traceback

try:
    import win32com.client
    from pythoncom import CoInitialize, CoUninitialize
except:
    print "Trouble with Python/COM:"
    traceback.print_exc()

QUB_Colors = [0, 0x0000ff, 0x800000, 0x008000, 0x00ffff, 0xff00ff, 0x00ff00, 0xff0000, 0xffff00, 0x800080]

def MapOverChildren(func, node):
    result = []
    ch = node.child
    while not ch.isNull:
        result.append( func(ch) )
        ch = ch.sibling
    return result

# ------- 'Select': the Select button in the Select tab

def SelectAllUsingColor(res, col):
    res.doUpdates(None) # turn off auto-display-updates
    for i in range(0, res.segmentCount):
        res.setSegmentColor(i, col)
    res.doUpdates(1) # turn on auto-display-updates

def SelectAll():
    res = QUB.ResultsView.currResults
    col = res.chosenColor
    SelectAllUsingColor(res,col)

def SelectNone():
    res = QUB.ResultsView.currResults
    SelectAllUsingColor(res, 0)

Commands.Register('Select', 'All', SelectAll)
Commands.Register('Select', 'None', SelectNone)

def prod(lst):
    res = 1.0
    for x in lst:
        res *= x
    return res

def sum(lst):
    res = 0.0
    for x in lst:
        res += x
    return res

def getUsedNonzeroColors(result):
    colors = result.usedColors
    try:
        colors.remove(0)
    except:
        pass
    return colors

def max_index(lst):
    mxi = 0
    for i in range(1,len(lst)):
        if lst[i] > lst[mxi]:
            mxi = i
    return mxi

def TrimSegs(res, segs):    
    segcolor = map( res.segmentColor, segs )
    numSD = res.numSD
    
    for var in map( lambda varname: res[varname], res.visibleVars ):
        for segi in range(0, len(segs)):
            seg = segs[ segi ]
            col = segcolor[ segi ]
            if col > 0:
                if fabs(var.mean(col) - var[seg]) > (numSD * var.sd(col)):
                    segcolor[segi] = 0

    res.doUpdates(0)
    for i in range(0, len(segs)):
        res.setSegmentColor( segs[i], segcolor[i] )
    res.doUpdates(1)
    
def Trim():
    res = QUB.ResultsView.currResults
    segs = range(0, res.segmentCount)
    TrimSegs(res, segs)

Commands.Register('Select', 'Trim', Trim)

def MakeSelLists():
    res = QUB.ResultsView.currResults
    for color in res.usedColors:
        NewSelListOfColor(color)    # below, in SelectColor section

Commands.Register('Select', 'Make selection lists...', MakeSelLists)        

# --------- SelectColor: right-click a color in Results::Select

def TrimColor():
    res = QUB.ResultsView.currResults
    segs = res.segmentsOfColor( res.chosenColor )
    TrimSegs(res, segs)

Commands.Register('SelectColor', 'Trim', TrimColor)

Commands.Register('SelectColor', 'Select All', SelectAll)

def SelectColorBetweenLines( onlyBlackOnes = 0 ):
    res = QUB.ResultsView.currResults
    col = res.chosenColor
    numSD = res.numSD
    segs = range(0, res.segmentCount)
    segcolor = map( lambda i: res.segmentColor(i), segs )

    for var in map( lambda varname: res[varname], res.visibleVars ):
        for seg in segs:
            if (segcolor[seg] == 0) or (not onlyBlackOnes):
                if fabs(var.mean(col) - var[seg]) <= (numSD * var.sd(col)):
                    segcolor[seg] = col

    res.doUpdates(0)
    for i in segs:
        res.setSegmentColor( i, segcolor[i] )
    res.doUpdates(1)

def SelectColorBlackBetweenLines():
    SelectColorBetweenLines(1)

Commands.Register('SelectColor', 'Select All Between Lines', SelectColorBetweenLines)
Commands.Register('SelectColor', 'Select Black Between Lines', SelectColorBlackBetweenLines)

def ExtractColor():
    res = QUB.ResultsView.currResults
    newres = res.node.clone()
    newres.name = os.path.splitext( os.path.split( res.fileName )[1] )[0] + 'x'

    chosenSegs = res.segmentsOfColor( res.chosenColor )
    segs = []
    
    ds = newres.find('DataSet')
    while not ds.isNull:
        seg = ds.find('Segment')
        while not seg.isNull:
            segs.append(seg)
            seg = seg.nextSameName()
        ds = ds.nextSameName()

    for seg in range(0, res.segmentCount):
        if not (seg in chosenSegs):
            segs[ seg ].parent.remove( segs[ seg ] )

    QUB.ResultsView.addResults( newres )

Commands.Register('SelectColor', 'Extract Segments', ExtractColor)

def RemoveColor():
    res = QUB.ResultsView.currResults
    newres = res.node.clone()
    newres.name = res.fileName

    chosenSegs = res.segmentsOfColor( res.chosenColor )
    segs = []
    
    ds = newres.find('DataSet')
    while not ds.isNull:
        seg = ds.find('Segment')
        while not seg.isNull:
            segs.append(seg)
            seg = seg.nextSameName()
        ds = ds.nextSameName()

    for seg in range(0, res.segmentCount):
        if seg in chosenSegs:
            segs[ seg ].parent.remove( segs[ seg ] )

    QUB.ResultsView.addResults( newres ) # same name replaces

Commands.Register('SelectColor', 'Remove Segments', RemoveColor)


ColorNames = ['black', 'red', 'darkBlue', 'green', 'yellow', 'pink', 'lightGreen', 'blue', 'lightBlue', 'purple']

def GetDataByName( name ):
    dataView = QUB.DataView
    for i in range(len(dataView)):
        data = dataView[i]
        if data.fileName == name:
            return data

def NewSelListOfColor(color, defName=None):
    res = QUB.ResultsView.currResults

    # prompt for list name
    colorName = ColorNames[color]
    name = defName
    if not name:
        name = '%s:%s' % (str(res.node['ResultType'].data), colorName)
    response = QUB.Dialog('New selection list with %s segments' % colorName, ['OK', 'Cancel'],
                          [DialogText('listName', 'List name: ', name)])
    if response['button'] == 'Cancel':
        return
    listName = response['listName']

    frontData = QUB.DataView.currData
    
    # clear existing lists by that name
    ix = frontData.lists.indexOf(listName)
    if ix >= 0:
        frontData.lists.delete(ix)
        
    ds = res.node.find('DataSet')
    while ds:
        currData = GetDataByName( str(ds['FileName'].data) )
        if currData:
            ix = currData.lists.indexOf(listName)
            if ix >= 0:
                currData.lists.delete(ix)
        ds = ds.nextSameName()

    # add each sel-of-color to the list in the appropriate data file    
    currDataName = '<none>'
    currData = None
    currList = 'None'
    for seg in [res.segment(i) for i in res.segmentsOfColor( color )]:
        if str(seg.parent['FileName'].data) <> currDataName:
            currDataName = str(seg.parent['FileName'].data)
            currData = GetDataByName( currDataName ) or frontData
            if currData:
                currList = currData.lists.byName( listName )
                currData.lists.choose( currData.lists.indexOf( listName ) )
            else:
                currList = 'None'
        if currList <> 'None':
            currList.add( seg.data[0], seg.data[1] )

def NewSelListOfChosenColor():
    NewSelListOfColor(QUB.ResultsView.currResults.chosenColor)
    
Commands.Register('SelectColor', 'Make Selection List...', NewSelListOfChosenColor)


# ------------ CriteriaAction: "Do..." on the Criteria tab

def CritToFunc(res, crit):
    critstr = str( crit[0].data )
    opstr = str( crit[1].data )
    op = lambda x,y: 1 # def: accept everything

    if critstr == 'Color':
        color = 0
        try:
            color = int(crit[2].data[0])
        except:
            pass

        if opstr == 'is':
            op = lambda x,y: x==y
        elif opstr == 'is not':
            op = lambda x,y: x!=y

        return lambda seg: op( res.segmentColor(seg), color )
    
    else:        
        var = res[ critstr ]
        if opstr == '<':
            op = lambda x,y: x<y
        elif opstr == '<=':
            op = lambda x,y: x<=y
        elif opstr == '>':
            op = lambda x,y: x>y
        elif opstr == '>=':
            op = lambda x,y: x>=y

        #val = 0
        #try:
        #    val = eval(str(crit[2].data))
        #except:
        #    pass
        
        valstr = str(crit[2].data)
        # global search/replace:
        #  "mean" -> var.mean( res.segmentColor(seg) )
        #  "std"  -> var.sd( res.segmentColor(seg) )
        valstr = re.sub('(^|[^.])mean', r'\1var.mean( res.segmentColor(seg) )', valstr)
        valstr = re.sub('(^|[^.])std', r'\1var.sd( res.segmentColor(seg) )', valstr)

        return lambda seg: (0 and res) or op( var[seg], eval(valstr) )

        # about that (0 and res) bit: (pretend it's not there)
        #     It appears that variables from nested scopes are
        #     available only if they are in the text of the lambda
        #     (so variables mentioned only in valstr might be optimized away)
        #     Thus, we need to mention res somewhere.
        #     The trick chosen relies on short-circuit evaluation and
        #     the fact that (false and true) or expr == expr

def SegmentsMatchingCriteria(res):
    segs = range(0, res.segmentCount)[:]

    for crit in MapOverChildren( lambda critNode: CritToFunc(res, critNode), res.criteria ):
        disqualified = []
        for seg in segs:
            if not crit(seg):
                disqualified.append(seg)
        for seg in disqualified:
            segs.remove(seg)

    return segs

def SelectOnCriteria():
    res = QUB.ResultsView.currResults
    color = res.chosenColor

    res.doUpdates(0)
    for seg in SegmentsMatchingCriteria(res):
        res.setSegmentColor( seg, color )
    res.doUpdates(1)

Commands.Register('CriteriaAction', 'Select', SelectOnCriteria)

def UnselectOnCriteria():
    res = QUB.ResultsView.currResults

    res.doUpdates(0)
    for seg in SegmentsMatchingCriteria(res):
        res.setSegmentColor( seg, 0 )
    res.doUpdates(1)

Commands.Register('CriteriaAction', 'Unselect', UnselectOnCriteria)


# ------------------------------- Extract Select Chart to Excel ---------------------------

ColorMarkers = [0x8,        # xlMarkerStyleCircle
                0x3,        # xlMarkerStyleTriangle
                0x2,        # xlMarkerStyleDiamond
                0x9,        # xlMarkerStylePlus
                0x8,        # xlMarkerStyleCircle
                0x5,        # xlMarkerStyleStar
                0x1,        # xlMarkerStyleSquare
                -4168,      # 0xffffefb8, # xlMarkerStyleX
                0x3,        # xlMarkerStyleTriangle (repeat)
                0x1]        # xlMarkerStyleSquare (repeat)
#               0xffffefea, # xlMarkerStyleDot (unused due to smallness and lack of color)
# 0xffffefed, # xlMarkerStyleDash

ColorMarkerSizes = [4,      # xlMarkerStyleDash
                    8,      # xlMarkerStyleTriangle
                    8,      # xlMarkerStyleDiamond
                    8,      # xlMarkerStylePlus
                    8,      # xlMarkerStyleCircle
                    8,      # xlMarkerStyleStar
                    8,      # xlMarkerStyleSquare
                    8,      # xlMarkerStyleX
                    10,      # xlMarkerStyleTriangle (repeat)
                    10]      # xlMarkerStyleSquare (repeat)

MarkerColors = QUB_Colors[:]
MarkerColors[0] = 0xffffff

def ExtractChartToExcel(varname, byStartTime=0):
    res = QUB.ResultsView.currResults
    var = res[varname]

    segNodes = []
    segNode = res.node['DataSet'].find('Segment')
    while not segNode.isNull:
        segNodes.append( segNode )
        segNode = segNode.nextSameName()

    startTimes = res['start']
    if not startTimes:
        startTimes = map( lambda seg: 0.0, range(0, res.segmentCount) )

    units = var.units

    xvals = 1 # column 1: seg index
    xtitle = ''
    if byStartTime:
        xvals = 2
        xtitle = 'start (sec)'

    usedColors = res.usedColors

    #CoInitialize() -- let WordReport.py do this on startup
    try:
        excel = win32com.client.Dispatch("Excel.Application")
        book = excel.Workbooks.Add(-4167) # XLWBatWorksheet
        sheet = book.Worksheets.Item(1)
        excel.Visible = 1
        cells = sheet.Cells

        sheet.ChartObjects().Add(50,14*(1+len(usedColors)),500,280)
        chart = sheet.ChartObjects(1).Chart
        chart.ChartType = -4169 # 0xffffefb7 # xlXYScatter = 0xffffefb7
        while chart.SeriesCollection().Count > 0:
            chart.SeriesCollection(1).Delete

        cells(1,1).Value = 'Segment'
        cells(1,2).Value = 'Start'
        cells(1,3).Value = 'Bounds'
        cells(1,4).Value = varname
        cells(1,5).Value = 'Color'
        cells(1,6).Value = 'RGBColor'

        cells(1,8).Value = 'Color'
        cells(1,9).Value = 'Mean'
        cells(1,10).Value = 'Std'

        curline = 2
        for color_i in range(0, len(usedColors)):
            color = usedColors[ color_i ]
            cells(color_i+2,8).Value = color
            cells(color_i+2,9).Value = var.mean(color)
            cells(color_i+2,10).Value = var.sd(color)

            colorFirstLine = curline
            for seg in res.segmentsOfColor(color):
                cells(curline, 1).Value = seg+1
                cells(curline, 2).Value = startTimes[seg] * 1.0e-3
                cells(curline, 3).Value = str(segNodes[seg].data[0]) + ' - ' + str(segNodes[seg].data[1])
                cells(curline, 4).Value = var[seg]
                cells(curline, 5).Value = color
                cells(curline, 6).Value = QUB_Colors[ res.segmentColor(seg) ]

                curline = curline + 1

            chart.SeriesCollection().Add( sheet.Range( sheet.Cells(colorFirstLine,4), sheet.Cells(curline-1,4) ), 2, 0, 0, 0)
            dots = chart.SeriesCollection(color_i+1)
            dots.XValues = sheet.Range( sheet.Cells(colorFirstLine,xvals), sheet.Cells(curline-1,xvals) )
            dots.Name = ('%.3f' % var.mean(color)) + ' ~ ' + ('%.3f' % var.sd(color))
            dots.MarkerForegroundColor = QUB_Colors[0]
            dots.MarkerBackgroundColor = MarkerColors[ color ]
            dots.MarkerStyle           = ColorMarkers[ color ]
            dots.MarkerSize            = ColorMarkerSizes[ color ]

        yaxis = chart.Axes(2)
        yaxis.HasTitle = 1
        yaxis.AxisTitle.Text = units
        yaxis.CrossesAt = yaxis.MinimumScale

        xaxis = chart.Axes(1)
        xaxis.HasTitle = 1
        xaxis.AxisTitle.Caption = xtitle

        chart.HasLegend = 1
        if 0 in usedColors:
            chart.Legend.LegendEntries(1).Delete()
        
        chart.PlotArea.Interior.ColorIndex = 0
        chart.HasTitle = 1
        chart.ChartTitle.Text = varname
        
        sheet.Range( cells(1,8), cells(len(usedColors)+1,10) ).Borders.LineStyle = 1 # xlContinuous
    finally:
        pass
        #CoUninitialize()

def Extract2ChartToExcel(varname1, varname2):
    res = QUB.ResultsView.currResults
    var1 = res[varname1]
    var2 = res[varname2]

    segNodes = []
    segNode = res.node['DataSet'].find('Segment')
    while not segNode.isNull:
        segNodes.append( segNode )
        segNode = segNode.nextSameName()

    units1 = var1.units
    units2 = var2.units

    usedColors = res.usedColors

    #CoInitialize()
    try:
        excel = win32com.client.Dispatch("Excel.Application")
        book = excel.Workbooks.Add(-4167) # XLWBatWorksheet
        sheet = book.Worksheets.Item(1)
        excel.Visible = 1
        cells = sheet.Cells

        sheet.ChartObjects().Add(50,14*(1+len(usedColors)),500,280)
        chart = sheet.ChartObjects(1).Chart
        chart.ChartType = -4169 # 0xffffefb7 # xlXYScatter = 0xffffefb7
        while chart.SeriesCollection().Count > 0:
            chart.SeriesCollection(1).Delete

        cells(1,1).Value = 'Segment'
        cells(1,2).Value = 'Bounds'
        cells(1,3).Value = '%s [%s]' % (varname1, units1)
        cells(1,4).Value = '%s [%s]' % (varname2, units2)
        cells(1,5).Value = 'Color'
        cells(1,6).Value = 'RGBColor'

        cells(1,8).Value = 'Color'
        cells(1,9).Value = 'Mean X'
        cells(1,10).Value = 'Std X'
        cells(1,11).Value = 'Mean Y'
        cells(1,12).Value = 'Std Y'

        curline = 2
        for color_i in range(0, len(usedColors)):
            color = usedColors[ color_i ]
            cells(color_i+2,8).Value = color
            cells(color_i+2,9).Value = var1.mean(color)
            cells(color_i+2,10).Value = var1.sd(color)
            cells(color_i+2,11).Value = var2.mean(color)
            cells(color_i+2,12).Value = var2.sd(color)

            colorFirstLine = curline
            for seg in res.segmentsOfColor(color):
                cells(curline, 1).Value = seg+1
                cells(curline, 2).Value = str(segNodes[seg].data[0]) + ' - ' + str(segNodes[seg].data[1])
                cells(curline, 3).Value = var1[seg]
                cells(curline, 4).Value = var2[seg]
                cells(curline, 5).Value = color
                cells(curline, 6).Value = QUB_Colors[ res.segmentColor(seg) ]

                curline = curline + 1

            chart.SeriesCollection().Add( sheet.Range( sheet.Cells(colorFirstLine,4), sheet.Cells(curline-1,4) ), 2, 0, 0, 0)
            dots = chart.SeriesCollection(color_i+1)
            dots.XValues = sheet.Range( sheet.Cells(colorFirstLine,3), sheet.Cells(curline-1,3) )
            dots.MarkerForegroundColor = QUB_Colors[0]
            dots.MarkerBackgroundColor = MarkerColors[ color ]
            dots.MarkerStyle           = ColorMarkers[ color ]
            dots.MarkerSize            = ColorMarkerSizes[ color ]

        xaxis = chart.Axes(1)
        xaxis.HasTitle = 1
        xaxis.AxisTitle.Text = cells(1,3).Value

        yaxis = chart.Axes(2)
        yaxis.HasTitle = 1
        yaxis.AxisTitle.Text = cells(1,4).Value
        yaxis.CrossesAt = yaxis.MinimumScale

        chart.HasLegend = 0
        
        chart.PlotArea.Interior.ColorIndex = 0
        chart.HasTitle = 1
        chart.ChartTitle.Text = varname1 + ' vs. ' + varname2
        
        sheet.Range( cells(1,8), cells(len(usedColors)+1,12) ).Borders.LineStyle = 1 # xlContinuous
    finally:
        pass
        #CoUninitialize()
